﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using GroupLab.Networking; // You need to include GroupLab.Netowrking in the solution references

namespace SharedDictionaryExample
{
    /// <summary>
    /// Interaction logic illustrating the Shared Dictionary
    /// </summary>
    public partial class Window1 : Window
    {
        SharedDictionary shDict = new SharedDictionary(); // The shared dictionary we will connect to
        Subscription subscription = new Subscription();   // The data in the dictionray we will subscribe to
        const string PATTERN = "/increment";              // The regular expression pattern we will set, get and subscribe to in the shared dictionary

        // All UI elements can be only changed if they are executed on the thread
        // that created the UI element. The delegate lets us do this. Note that this is 
        // a standard C# thing; it is not particular to the shared dictionary. 
        // This examples uses delegates where we explicitly specify the arguments.
        private delegate void TranscriptAddLineDelegate(TextBox textBox, string line);  //Add a line to the control
        private delegate void ControlIsEnabledDelegate(Control control, Boolean state); //Change the enabled state of the button
        private delegate void CheckBoxStateDelegate(CheckBox checkBox, Boolean state); //Change the enabled state of the button

        public Window1()
        {
            InitializeComponent();
            ControlEnabledState(this.btnIncrement, false);
        }

        #region Cleanly Configure or End the Shared Dictionary and Subscriptions
 
        // Configure up the shared dictionary
        private void ConfigureDictionary()
        {
            // Look for the server at the URL provided in the text box
            this.shDict.Url = textBoxShDictURL.Text;       

            // These are all the significant events thrown by a shared dictionary
            this.shDict.Connected += new EventHandler(shDict_Connected); // the client is connected to the server and contents are downloaded, but subscriptions have not yet been triggered
            this.shDict.Opened += new EventHandler(shDict_Opened); // Occurs when the connection to the shDictionary is opened
            this.shDict.Closed += new EventHandler(shDict_Closed);
            this.shDict.Disconnected += new EventHandler<DisconnectedEventArgs>(shDict_Disconnected);                             
        }


        // Because we are regularly opening and closing  the shared dictionary in this test example, 
        // we should remove its event handlers
        private void EndDictionary()
        {
            this.shDict.Close(); // Do this first so we still get the various  events associated with clossing the dictionary
            this.shDict.Connected -= new EventHandler(shDict_Connected); 
            this.shDict.Opened -= new EventHandler(shDict_Opened); 
            this.shDict.Closed -= new EventHandler(shDict_Closed);
            this.shDict.Disconnected -= new EventHandler<DisconnectedEventArgs>(shDict_Disconnected);
        }


        // Initialize the subscription
        // We put this inside a begin/end init to have it all processed together
        private void ConfigureSubscriptions()
        {
            this.subscription.BeginInit();
            this.subscription.Dictionary = this.shDict;  // The dictionary associated with this subscription
            this.subscription.Pattern = PATTERN;         // Register the particular pattern we are interested in 
            this.subscription.Notified += new SubscriptionEventHandler(subscription_Notified);  // This  handler is invoked when the pattern's value is added/changed/deleted
            this.subscription.EndInit();
        }

        // End the subscription
        // If we are regularly closing and opening the shared dictionary (as in this test example) we should probably do this
        private void EndSubscriptions()
        {
            this.subscription.BeginInit();
            this.subscription.Notified -= new SubscriptionEventHandler(subscription_Notified); // register the event handler
            this.subscription.EndInit();
        }

        #endregion

        #region User Interface Delegates
        // Add a line to the textbox
        private void TranscriptAddLine(TextBox tbox, string str)
        {
            this.Dispatcher.BeginInvoke(System.Windows.Threading.DispatcherPriority.Normal,
               (TranscriptAddLineDelegate)delegate(TextBox textBox, string line)
               {
                   textBox.Text = line + "\n" + textBox.Text;
               },
               tbox, str);
        }

        // Enable or Disable the user control
        void ControlEnabledState(object control, Boolean state)
        {
            // Change the button enabled to true
            this.Dispatcher.BeginInvoke(System.Windows.Threading.DispatcherPriority.Normal,
                (ControlIsEnabledDelegate)delegate(Control ctl, Boolean value)
                {
                    ctl.IsEnabled = value;
                },
                control, state);
        }

        void CheckBoxState(CheckBox checkBox, Boolean state)
        {
            // Change the check button state
            this.Dispatcher.BeginInvoke(System.Windows.Threading.DispatcherPriority.Normal,
                (CheckBoxStateDelegate)delegate(CheckBox cBox, Boolean newState)
                {
                    cBox.IsChecked = newState;
                },
                checkBox, state);
        }
        #endregion 

        #region Shared Dictionary event handlers

        // Occurs when a shared dictionary client has successfully connected to the
        // server and downloaded its initial copy of the contents of the shared dictionary,
        // but before any subscriptions held a-priori are notified.
        void shDict_Connected(object sender, EventArgs e)
        {
            TranscriptAddLine(this.textBoxTranscript, "Connected");
        }
        
        // Occurs when the primary connection to the shared dictionary server is opened,
        // and after initial subscription notifications have been invoked.
        void shDict_Opened(object sender, EventArgs e)
        {
            TranscriptAddLine(this.textBoxTranscript, "Opened");
            ControlEnabledState(this.btnIncrement, true);
            ControlEnabledState(this.textBoxShDictURL, false);
        }

        //     Occurs when the shared dictionary connection is broken. This event is raised
        //     immediately to describe the reason for the break and to provide an opportunity
        //     for the connection to be automatically re-established. The first attempt
        //     to re-establish happens immediately. If that attempt fails, the client backs
        //     off and waits about 20 seconds before trying again. Around four such retry
        //     attempts are made.
        //     If all four retry attempts fail, the Disconnected event is raised again to
        //     get confirmation that retrying should continue. If one of the retry attempts
        //     is successful, then the normal logon sequence will occur (the next events
        //     raised by the shared dictionary will be the GroupLab.Networking.SharedDictionary.Connected
        //     and GroupLab.Networking.SharedDictionary.Opened events).
        //     If retrying is not indicated, the next event to be raised by the shared dictionary
        //     will be the GroupLab.Networking.SharedDictionary.Closed event, which signals
        //     that the tear down of the connection is complete and that it will not be
        //     re-established without another call to the GroupLab.Networking.SharedDictionary.Open()
        //     method.
        void shDict_Disconnected(object sender, DisconnectedEventArgs e)
        {
            TranscriptAddLine(this.textBoxTranscript, "Disconnected: (" + e.Message + "), retry# " + "what happened to e.RetryCount ???");
            e.Retry = true;
        }

        // Occurs when the primary connection to the shared dictionary server is closed
        // and indicates that the connection will not be automatically restablished.
        void shDict_Closed(object sender, EventArgs e)
        {
            TranscriptAddLine(this.textBoxTranscript, "Closed");
            ControlEnabledState(this.btnIncrement, false);
            ControlEnabledState(this.textBoxShDictURL, false);
            CheckBoxState(this.checkBoxConnect, false);
        }
        #endregion

        #region Subscription Event Handlers
        // This event handler is called whenever the subsription pattern is altered. 
        // The reasons are also included,and different actions can be taken depending on that reason.
        void subscription_Notified(object sender, SubscriptionEventArgs e)
        {
            // Compose a message based on the reason, the key and the value from the dictionary
            switch (e.Reason)
            {
	            case SubscriptionNotification.Add:
                    TranscriptAddLine(this.textBoxTranscript, "Notified (Added): " + e.Path + ": "  + e.Value.ToString ());
                    break;
                case SubscriptionNotification.AttributeChange:
                    TranscriptAddLine(this.textBoxTranscript, "Notified (Attribute): " + e.Path + ": " + e.Value.ToString ());
                    break;
                case SubscriptionNotification.Change:
                    TranscriptAddLine(this.textBoxTranscript, "Notified (Changed): " + e.Path + ": "  + e.Value.ToString ());
                    break;
                case SubscriptionNotification.Remove:
                    TranscriptAddLine(this.textBoxTranscript, "Notified (Removed): " + e.Path);
                    break;
                default:
                    break;
            }
        }
        #endregion

        #region Interface Event Handlers
        private void btnIncrement_Click(object sender, RoutedEventArgs e)
        {
            if (shDict.Status == ConnectionStatus.Opened)
            {
                //check if we have an entry in the dictionary that matches the pattern
                if (!this.shDict.Contains(PATTERN))
                {
                    this.shDict[PATTERN] = 0; // then we create one and set its value to 0
                }
                else
                {
                    // increment the existing value
                    this.shDict[PATTERN] = (int)this.shDict.GetEntry(PATTERN).Value + 1;
                }
            }
        }

        // Connect to the shared dictionary
        private void checkBoxConnect_Checked(object sender, RoutedEventArgs e)
        {
            // Make sure that the dictionary isn't already opened before trying to open it
            if (shDict.Status != ConnectionStatus.Opened)
            {
                ConfigureDictionary();     // Initialize the connection to the shared dictionary
                ConfigureSubscriptions();  // Initialize the subscriptions to data in the shared dictionary
                this.shDict.Open();         // Now connect to the shared dictionary
            }
        }

        private void checkBoxConnect_Unchecked(object sender, RoutedEventArgs e)
        {
            // Make sure that the dictionary isn't already opened before trying to open it
            if (shDict.Status != ConnectionStatus.Closed)
            {
                EndSubscriptions(); 
                EndDictionary();
            }
        }
        #endregion
    }
}
